@@ -11,4 +11,6 @@
 #undef HAVE_XFT
 #undef HAVE_SM
 #undef HAVE_XINERAMA
+#undef HAVE_SOLARIS_XINERAMA
+#undef HAVE_XFREE_XINERAMA
 #undef HAVE_SHAPE
