@@ -129,7 +129,7 @@ struct _MetaDisplay
   Atom atom_wm_client_machine;
   Atom atom_net_wm_state_fullscreen;
   Atom atom_net_workarea;
-  Atom atom_net_show_desktop;
+  Atom atom_net_showing_desktop;
   Atom atom_net_desktop_layout;
   Atom atom_manager;
   Atom atom_targets;
@@ -147,12 +147,18 @@ struct _MetaDisplay
   Atom atom_net_wm_action_maximize_vert;
   Atom atom_net_wm_action_change_desktop;
   Atom atom_net_wm_action_close;
+  Atom atom_net_wm_state_above;
+  Atom atom_net_wm_state_below;
   
   /* This is the actual window from focus events,
    * not the one we last set
    */
   MetaWindow *focus_window;
 
+  /* window we are expecting a FocusIn event for
+   */
+  MetaWindow *expected_focus_window;
+
   /* Most recently focused list. Always contains all
    * live windows.
    */
@@ -205,14 +211,18 @@ struct _MetaDisplay
   MetaWindow *grab_window;
   Window      grab_xwindow;
   int         grab_button;
-  int         grab_root_x;
-  int         grab_root_y;
+  int         grab_initial_root_x;
+  int         grab_initial_root_y;
+  int         grab_current_root_x;
+  int         grab_current_root_y;
   gulong      grab_mask;
   guint       grab_have_pointer : 1;
   guint       grab_have_keyboard : 1;
   MetaRectangle grab_initial_window_pos;
+  MetaRectangle grab_current_window_pos;
   MetaResizePopup *grab_resize_popup;
-
+  GTimeVal    grab_last_moveresize_time;
+  
 
   /* Keybindings stuff */
   MetaKeyBinding *screen_bindings;
@@ -236,6 +246,9 @@ struct _MetaDisplay
 
   /* Closing down the display */
   int closing;
+
+  /* Managed by group.c */
+  GHashTable *groups_by_leader;
 };
 
 gboolean      meta_display_open                (const char  *name);
@@ -281,6 +294,13 @@ MetaWorkspace* meta_display_get_workspace_by_screen_index (MetaDisplay   *displa
 Cursor         meta_display_create_x_cursor (MetaDisplay *display,
                                              MetaCursor   cursor);
 
+void     meta_display_set_grab_op_cursor (MetaDisplay *display,
+                                          MetaScreen  *screen,
+                                          MetaGrabOp   op,
+                                          gboolean     change_pointer,
+                                          Window       grab_xwindow,
+                                          Time         timestamp);
+
 gboolean meta_display_begin_grab_op (MetaDisplay *display,
                                      MetaScreen  *screen,
                                      MetaWindow  *window,
@@ -351,6 +371,11 @@ MetaWindow* meta_display_get_tab_next (MetaDisplay   *display,
                                        MetaWindow    *window,
                                        gboolean       backward);
 
+MetaWindow* meta_display_get_tab_current (MetaDisplay   *display,
+                                          MetaTabList    type,
+                                          MetaScreen    *screen,
+                                          MetaWorkspace *workspace);
+
 int meta_resize_gravity_from_grab_op (MetaGrabOp op);
 
 gboolean meta_rectangle_intersect (MetaRectangle *src1,
