@@ -60,10 +60,21 @@ log_handler (const gchar   *log_domain,
 static void
 usage (void)
 {
-  g_print (_("metacity [--disable-sm] [--sm-save-file=FILENAME] [--display=DISPLAY] [--replace]\n"));
+  g_print (_("metacity [--disable-sm] [--sm-save-file=FILENAME] [--display=DISPLAY] [--replace] [--version]\n"));
   exit (1);
 }
 
+static void
+version (void)
+{
+  g_print (_("metacity %s\n"
+             "Copyright (C) 2001-2002 Havoc Pennington, Red Hat, Inc., and others\n"
+             "This is free software; see the source for copying conditions.\n"
+             "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
+           VERSION);
+  exit (0);
+}
+
 int
 main (int argc, char **argv)
 {
@@ -136,6 +147,8 @@ main (int argc, char **argv)
           strcmp (arg, "-h") == 0 ||
           strcmp (arg, "-?") == 0)
         usage ();
+      else if (strcmp (arg, "--version") == 0)
+        version ();
       else if (strcmp (arg, "--sm-disable") == 0)
         disable_sm = TRUE;
       else if (strcmp (arg, "--replace") == 0)
@@ -243,9 +256,19 @@ main (int argc, char **argv)
   meta_verbose ("Compiled without shape extension\n");
 #endif
 #ifdef HAVE_XINERAMA
-  meta_verbose ("Compiled with Xinerama extension\n");
+  meta_topic (META_DEBUG_XINERAMA, "Compiled with Xinerama extension\n");
 #else
-  meta_verbose ("Compiled without Xinerama extension\n");
+  meta_topic (META_DEBUG_XINERAMA, "Compiled without Xinerama extension\n");
+#endif
+#ifdef HAVE_XFREE_XINERAMA
+  meta_topic (META_DEBUG_XINERAMA, " (using XFree86 Xinerama)\n");
+#else
+  meta_topic (META_DEBUG_XINERAMA, " (not using XFree86 Xinerama)\n");
+#endif
+#ifdef HAVE_SOLARIS_XINERAMA
+  meta_topic (META_DEBUG_XINERAMA, " (using Solaris Xinerama)\n");
+#else
+  meta_topic (META_DEBUG_XINERAMA, " (not using Solaris Xinerama)\n");
 #endif
   
   /* Load prefs */
@@ -257,7 +280,7 @@ main (int argc, char **argv)
   /* must be after UI init so we can override GDK handlers */
   meta_errors_init ();
 
-#if 0
+#if 1
   g_log_set_handler (NULL,
                      G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
                      log_handler, NULL);
