@@ -542,6 +542,7 @@ change_notify (GConfClient    *client,
   
  out:
   /* nothing */
+  return; /* AIX compiler wants something after a label like out: */
 }
 
 static gboolean
@@ -873,6 +874,8 @@ static MetaKeyPref window_bindings[] = {
   { META_KEYBINDING_WINDOW_MENU, 0, 0 },
   { META_KEYBINDING_TOGGLE_FULLSCREEN, 0, 0 },
   { META_KEYBINDING_TOGGLE_MAXIMIZE, 0, 0 },
+  { META_KEYBINDING_MAXIMIZE, 0, 0 },
+  { META_KEYBINDING_UNMAXIMIZE, 0, 0 },
   { META_KEYBINDING_TOGGLE_SHADE, 0, 0 },
   { META_KEYBINDING_MINIMIZE, 0, 0 },
   { META_KEYBINDING_CLOSE, 0, 0 },
@@ -896,6 +899,8 @@ static MetaKeyPref window_bindings[] = {
   { META_KEYBINDING_MOVE_WORKSPACE_UP, 0, 0 },
   { META_KEYBINDING_MOVE_WORKSPACE_DOWN, 0, 0 },
   { META_KEYBINDING_RAISE_OR_LOWER, 0, 0 },
+  { META_KEYBINDING_RAISE, 0, 0 },
+  { META_KEYBINDING_LOWER, 0, 0 },
   { NULL, 0, 0 }
 };
 
@@ -1183,3 +1188,25 @@ meta_prefs_get_keybinding_action (const char *name)
   return META_KEYBINDING_ACTION_NONE;
 }
 
+void
+meta_prefs_get_window_binding (const char          *name,
+                               unsigned int        *keysym,
+                               MetaVirtualModifier *modifiers)
+{
+  int i;
+
+  i = G_N_ELEMENTS (window_bindings) - 2; /* -2 for dummy entry at end */
+  while (i >= 0)
+    {
+      if (strcmp (window_bindings[i].name, name) == 0)
+        {
+          *keysym = window_bindings[i].keysym;
+          *modifiers = window_bindings[i].modifiers;
+          return;
+        }
+      
+      --i;
+    }
+
+  g_assert_not_reached ();
+}
