@@ -107,6 +107,8 @@ void meta_prefs_set_num_workspaces (int n_workspaces);
 #define META_KEYBINDING_WINDOW_MENU          "activate_window_menu"
 #define META_KEYBINDING_TOGGLE_FULLSCREEN    "toggle_fullscreen"
 #define META_KEYBINDING_TOGGLE_MAXIMIZE      "toggle_maximized"
+#define META_KEYBINDING_MAXIMIZE             "maximize"
+#define META_KEYBINDING_UNMAXIMIZE           "unmaximize"
 #define META_KEYBINDING_TOGGLE_SHADE         "toggle_shaded"
 #define META_KEYBINDING_MINIMIZE             "minimize"
 #define META_KEYBINDING_CLOSE                "close"
@@ -130,6 +132,8 @@ void meta_prefs_set_num_workspaces (int n_workspaces);
 #define META_KEYBINDING_MOVE_WORKSPACE_UP    "move_to_workspace_up"
 #define META_KEYBINDING_MOVE_WORKSPACE_DOWN  "move_to_workspace_down"
 #define META_KEYBINDING_RAISE_OR_LOWER       "raise_or_lower"
+#define META_KEYBINDING_RAISE                "raise"
+#define META_KEYBINDING_LOWER                "lower"
 
 typedef enum _MetaKeyBindingAction
 {
@@ -183,6 +187,10 @@ void meta_prefs_get_window_bindings (const MetaKeyPref **bindings,
 
 MetaKeyBindingAction meta_prefs_get_keybinding_action (const char *name);
 
+void meta_prefs_get_window_binding (const char          *name,
+                                    unsigned int        *keysym,
+                                    MetaVirtualModifier *modifiers);
+
 #endif
 
 
