@@ -944,6 +944,7 @@ save_state (void)
               char *res_class;
               char *res_name;
               char *role;
+              char *title;
 
               /* client id, class, name, role are not expected to be
                * in UTF-8 (I think they are in XPCS which is Latin-1?
@@ -957,6 +958,10 @@ save_state (void)
                 encode_text_as_utf8_markup (window->res_name) : NULL;
               role = window->role ?
                 encode_text_as_utf8_markup (window->role) : NULL;
+              if (window->title)
+                title = g_markup_escape_text (window->title, -1);
+              else
+                title = NULL;
               
               meta_topic (META_DEBUG_SM, "Saving session managed window %s, client ID '%s'\n",
                           window->desc, window->sm_client_id);
@@ -966,7 +971,7 @@ save_state (void)
                        sm_client_id,
                        res_class ? res_class : "",
                        res_name ? res_name : "",
-                       window->title ? window->title : "",
+                       title ? title : "",
                        role ? role : "",
                        window_type_to_string (window->type),
                        stack_position);
@@ -975,6 +980,7 @@ save_state (void)
               g_free (res_class);
               g_free (res_name);
               g_free (role);
+              g_free (title);
               
               /* Sticky */
               if (window->on_all_workspaces)
@@ -986,7 +992,14 @@ save_state (void)
 
               /* Maximized */
               if (window->maximized)
-                fputs ("    <maximized/>\n", outfile);
+		{
+                  fprintf (outfile,
+                           "    <maximized saved_x=\"%d\" saved_y=\"%d\" saved_width=\"%d\" saved_height=\"%d\"/>\n", 
+                           window->saved_rect.x,
+                           window->saved_rect.y,
+                           window->saved_rect.width,
+                           window->saved_rect.height);
+		}
               
               /* Workspaces we're on */
               {
@@ -1347,8 +1360,70 @@ start_element_handler  (GMarkupParseContext *context,
     }
   else if (strcmp (element_name, "maximized") == 0)
     {
+      int i;
+
+      i = 0;
       pd->info->maximized = TRUE;
       pd->info->maximized_set = TRUE;
+      while (attribute_names[i])
+        {
+          const char *name;
+          const char *val;
+
+          name = attribute_names[i];
+          val = attribute_values[i];
+
+          if (strcmp (name, "saved_x") == 0)
+            {
+              if (*val)
+                {
+                  pd->info->saved_rect.x = atoi (val);
+                  pd->info->saved_rect_set = TRUE;
+                }
+            }
+          else if (strcmp (name, "saved_y") == 0)
+            {
+              if (*val)
+                {
+                  pd->info->saved_rect.y = atoi (val);
+                  pd->info->saved_rect_set = TRUE;
+                }
+            }
+          else if (strcmp (name, "saved_width") == 0)
+            {
+              if (*val)
+                {
+                  pd->info->saved_rect.width = atoi (val);
+                  pd->info->saved_rect_set = TRUE;
+                }
+            }
+          else if (strcmp (name, "saved_height") == 0)
+            {
+              if (*val)
+                {
+                  pd->info->saved_rect.height = atoi (val);
+                  pd->info->saved_rect_set = TRUE;
+                }
+            }
+          else
+            {
+              g_set_error (error,
+                           G_MARKUP_ERROR,
+                           G_MARKUP_ERROR_UNKNOWN_ATTRIBUTE,
+                           _("Unknown attribute %s on <maximized> element"),
+                           name);
+              return;
+            }
+
+          ++i;
+        }
+
+      if (pd->info->saved_rect_set)
+        meta_topic (META_DEBUG_SM, "Saved unmaximized size %d,%d %dx%d \n",
+                    pd->info->saved_rect.x,
+                    pd->info->saved_rect.y,
+                    pd->info->saved_rect.width,
+                    pd->info->saved_rect.height);
     }  
   else if (strcmp (element_name, "geometry") == 0)
     {
