@@ -257,6 +257,28 @@ modal_dialog_cb (gpointer             callback_data,
   gtk_widget_destroy (dialog);
 }
 
+static void
+no_parent_dialog_cb (gpointer             callback_data,
+                     guint                callback_action,
+                     GtkWidget           *widget)
+{
+  GtkWidget *dialog;
+  
+  dialog = gtk_message_dialog_new (NULL,
+                                   GTK_DIALOG_DESTROY_WITH_PARENT,
+                                   GTK_MESSAGE_INFO,
+                                   GTK_BUTTONS_CLOSE,
+                                   "Here is a dialog with no transient parent");
+
+  /* Close dialog on user response */
+  g_signal_connect (G_OBJECT (dialog),
+                    "response",
+                    G_CALLBACK (gtk_widget_destroy),
+                    NULL);
+  
+  gtk_widget_show (dialog);
+}
+
 static void
 utility_cb (gpointer             callback_data,
             guint                callback_action,
@@ -613,6 +635,7 @@ static GtkItemFactoryEntry menu_items[] =
   { "/Windows/tearoff",       NULL,         NULL,                     0, "<Tearoff>" },
   { "/Windows/_Dialog",       "<control>d",  dialog_cb,               0, NULL },
   { "/Windows/_Modal dialog", NULL,          modal_dialog_cb,         0, NULL },
+  { "/Windows/_Parentless dialog", NULL,     no_parent_dialog_cb,     0, NULL },
   { "/Windows/_Utility",      "<control>u",  utility_cb,              0, NULL },
   { "/Windows/_Splashscreen", "<control>s",  splashscreen_cb,         0, NULL },
   { "/Windows/_Top dock",     NULL,          dock_cb,                 DOCK_TOP, NULL },
@@ -634,6 +657,17 @@ sleep_cb (GtkWidget *button,
   sleep (1000);
 }
 
+static void
+toggle_decorated_cb (GtkWidget *button,
+                     gpointer   data)
+{
+  GtkWidget *window;
+  window = gtk_widget_get_ancestor (button, GTK_TYPE_WINDOW);
+  if (window)
+    gtk_window_set_decorated (GTK_WINDOW (window),
+                              !gtk_window_get_decorated (GTK_WINDOW (window)));
+}
+
 static void
 clicked_toolbar_cb (GtkWidget *button,
                     gpointer   data)
@@ -779,6 +813,14 @@ do_appwindow (void)
                             window, /* user data for callback */
                             -1);  /* -1 means "append" */
 
+  gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
+                            GTK_STOCK_OPEN,
+                            "This is a demo button that toggles window decorations",
+                            NULL,
+                            G_CALLBACK (toggle_decorated_cb),
+                            window, /* user data for callback */
+                            -1);  /* -1 means "append" */
+  
   gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                             GTK_STOCK_QUIT,
                             "This is a demo button with a 'quit' icon",
