@@ -29,6 +29,8 @@
 #include <X11/Xutil.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
 
+typedef struct _MetaGroup MetaGroup;
+
 typedef void (*MetaWindowForeachFunc) (MetaWindow *window,
                                        void       *data);
 
@@ -160,6 +162,10 @@ struct _MetaWindow
   /* Computed whether to skip taskbar or not */
   guint skip_taskbar : 1;
   guint skip_pager : 1;
+
+  /* TRUE if client set these */
+  guint wm_state_above : 1;
+  guint wm_state_below : 1;
   
   /* this flag tracks receipt of focus_in focus_out and
    * determines whether we draw the focus
@@ -202,6 +208,9 @@ struct _MetaWindow
   guint has_struts : 1; 
   /* Struts are from the _WIN_HINTS do not cover deal */
   guint do_not_cover : 1;
+
+  /* Transient parent is a root window */
+  guint transient_parent_is_root_window : 1;
   
   /* Number of UnmapNotify that are caused by us, if
    * we get UnmapNotify with none pending then the client
@@ -247,11 +256,14 @@ struct _MetaWindow
   
   /* Managed by stack.c */
   MetaStackLayer layer;
-  MetaStackOp *stack_op;
-
+  int stack_position; /* see comment in stack.h */
+  
   /* Current dialog open for this window */
   int dialog_pid;
   int dialog_pipe;
+
+  /* maintained by group.c */
+  MetaGroup *cached_group;
 };
 
 #define META_WINDOW_ALLOWS_MOVE(w)     ((w)->has_move_func && !(w)->maximized && !(w)->fullscreen)
@@ -406,4 +418,16 @@ void meta_window_foreach_transient (MetaWindow            *window,
 gboolean meta_window_is_ancestor_of_transient (MetaWindow *window,
                                                MetaWindow *transient);
 
+gboolean meta_window_warp_pointer (MetaWindow *window,
+                                   MetaGrabOp  grab_op);
+
+void meta_window_begin_grab_op (MetaWindow *window,
+                                MetaGrabOp  op,
+                                Time        timestamp);
+
+void meta_window_update_resize_grab_op (MetaWindow *window,
+                                        gboolean    update_cursor);
+
+void meta_window_update_layer (MetaWindow *window);
+
 #endif
